#!/usr/bin/env python

# nuheclient.py v. 0.01
#
# Copyright (C) 2007 Tuomo Makinen (tjam@users.sourceforge.net)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
Nuhe node manager cli client version 0.01
Copyright (C) 2007 Tuomo Makinen

Usage:
  -i  				Interactive mode
  -a  		address		Node manager IP address
  -p  		port		Node manager TCP port
  -s  		name		Sensor name
  -c		filename	File which contains username, password and server:port
                                information (newline separated)

  Only one of the following options can be used at a time:

  --query      			Query sensors (shows list of registered sensors) 
  --status     			Sensor status
  --rulesets   			Sensor rulesets
  --sendrule   	ruleset	        Upload ruleset
  --fetchrule  	ruleset 	Download ruleset
  --sendconf   			Upload configuration file
  --fetchconf  			Download configuration file
  --slog       			Sensor logviewer
  --nlog       			Node manager logviewer
  --sendlog     string		Sends user specified string to sensor for alert inspection
  --freeze     			Freeze sensor
  --unfreeze   			Unfreeze sensor
  --restart    			Restart sensor
  --shutdown   			Shutdown sensor
  --remove                      Remove passive sensor entry from node manager

  --version    			Shows version
  --help       			Prints this message

  example:
  # nuheclient -i -a localhost:5403"""

import sys
import os
import string
import getopt
import getpass
import stat
import md5

from OpenSSL import crypto
    
import nmgrcomm

# Return opcodes from node manager
opcodes_r = [
      "102",
      "202",
      "204",
      "206",
      "212",
      "214",
      "216",
      "222",
      "224",
      "232",
      "242",
      "244",
      "262",
      "272",
      "282",
      "284",
      "292",
      "302"
    ]

# Globals
Interactive = False
Port = 5403
Addr = None
Sensor = None
Operation = None
Userdata = None
Confile = None

# Trusted certificates for verification purposes
if os.name == "posix":
    CA_CERTS = os.environ["HOME"] + "/.nuhecerts"
else:
    CA_CERTS = "nuhecerts"

def main():
    cn = Connector.open_connection (Addr, Port)
    if (cn[0] > 0):
        fatalerror("Error: can't open connection, " +
                   cn[1] + "\n")

    if (Connector.authenticate (Uname, Passwd) != None):
        if Interactive:
            do_menu()
        else:
            if F_dict.has_key(Operation):
                do_operation ()
            else:
                fatalerror("Error: I don't know what to do!\n")
    else:
       sys.stderr.write("Authentication failed\n")    

def get_name():
    while 1:
      try:
          uname = raw_input("Username: ")
      except KeyboardInterrupt:
          sys.stderr.write("\n")
          sys.exit(1)
      if uname == "":
          sys.stderr.write("Username can't be empty.\n")
      else:
          return uname

def get_passwd():
    while 1:
      try:
          return getpass.getpass()
      except KeyboardInterrupt:
          sys.stderr.write("\n")
          sys.exit(1)

def do_operation():
    if Operation not in ["--query", "--nlog"]:
        rval = Connector.select_sensor(Sensor)
        if (check_message(rval, False) > 0):
            sys.exit(1)

    if Operation == "--fetchrule" or \
       Operation == "--sendlog":
        rval = F_dict[Operation](Userdata)

    elif Operation == "--sendrule":
        rls = read_input_str("File to upload: ")
        if not rls:
            return
        content = open_file(rls)
        if content == None:
            sys.exit(1)
        else:
            rval = F_dict[Operation](Userdata, content)

    elif Operation == "--sendconf":
        rls = read_input_str("File to upload: ")
        if not rls:
            return
        content = open_file(rls)
        if content == None:
            sys.exit(1)
        else:
            rval = F_dict[Operation](content)

    else:
        if (Operation == "--slog" or
            Operation == "--nlog"):
            do_logviewer(F_dict[Operation])
            return
        else:
            rval = F_dict[Operation]()

    check_message(rval)
     
def do_menu():
    """1) Fetch sensor list\n2) Select sensor\n3) Nodemgr logviewer\n\nQ)uit"""

    sensor_s = 1
    sensor_n = ''

    print "Nodemanager: " + sys.argv[1] + \
          "\n\n" + do_menu.__doc__
    print "?) Help"

    while True:
        cmd = read_input_str("# ")
        if not cmd:
            Connector.logout()
            sys.exit(0)

        if cmd == "1":
            rval = Connector.query_sensors ()
            check_message (rval)
        elif cmd == "2":
            sname = read_input_str("sensor name: ")
            if not sname:
                continue
            rval = Connector.select_sensor(sname)
            if (check_message(rval) < 1):
	        sensor_s = 0
		sensor_n = sname
                do_sensor (sname)
        elif cmd == "3":
            do_logviewer (Connector.read_nodemgr_log)
	elif cmd == "s" or cmd == "S":
	    if sensor_s < 1:
	        do_sensor(sensor_n)
            else:
	        print "Unknown command"
        elif cmd == "q" or cmd == "Q":
            Connector.logout()
            sys.exit(0)
        elif cmd == "?":
	    print "Nodemanager: " + sys.argv[1] + \
	          "\n\n" + do_menu.__doc__
            if sensor_s < 1:
	        print "S)ensor menu"
            print "?) Help"
        else:
            print "Unknown command"

def do_sensor(sname):
    """1)  Sensor status\n2)  Sensor rulesets\n3)  Download ruleset\n4)  Upload ruleset\n\
5)  Download configuration file\n6)  Upload configuration file\n7)  Sensor logviewer\n\
8)  Nodemgr logviewer\n9)  Send logged data\n10) Remove sensor\n\nF)reeze sensor\nU)nfreeze sensor\n\
R)estart sensor\nS)hutdown sensor\n\nQ)uit\nP)revious menu\n?) Help"""

    print "Sensor: " + sname + \
          "\n\n" + do_sensor.__doc__

    while True:
    
        cmd = read_input_str("# ")
        if not cmd:
            Connector.logout()
            sys.exit(0)
          
        if cmd == "1":
            rval = Connector.query_sensor_status()
        elif cmd == "2":
            rval = Connector.query_rulesets()
        elif cmd == "3":
	    rname = read_input_str("Ruleset name: ")
	    if not rname:
	        continue
            rval = Connector.download_ruleset(rname)
            check_message(rval)
	    if (read_input_str("Save to file (y/n) ?") in ["y", "Y"]):
	        fname = read_input_str("Filename: ")
		if not fname:
		    continue
                save_file(fname, rval[4:])
            continue
        elif cmd == "4":
	    rlset = read_input_str ("Ruleset name: ")
	    if not rlset:
	        continue
	    fname = read_input_str ("File to upload: ")
	    if not fname:
	        continue
            content = open_file (fname)
            if content != None:
                rval = Connector.upload_ruleset(rlset, content)
            continue                                      
        elif cmd == "5":
            rval = Connector.download_configuration()
            check_message(rval)
	    if (read_input_str("Save to file (y/n) ? ") in ["y", "Y"]):
	        fname = read_input_str("Filename: ")
		if not fname:
		    continue
                save_file(fname, rval[4:])
            continue
        elif cmd == "6":
	    fname = read_input_str("File to upload: ")
	    if not fname:
	        continue
            content = open_file (fname)
            if content != None:
                rval = Connector.upload_configuration(content)
            continue
        elif cmd == "7":
            do_logviewer (Connector.read_sensor_log)
            continue 
        elif cmd == "8":
            do_logviewer (Connector.read_nodemgr_log)
            continue
        elif cmd == "9":
             ldata = read_input_str(">> ")
             if not ldata:
                 continue
             rval = Connector.send_logged_data(ldata)
        elif cmd == "10":
             rval = Connector.remove_sensor()
        elif cmd == "F" or cmd == "f":
            rval = Connector.freeze_sensor()
        elif cmd == "U" or cmd == "u":
            rval = Connector.unfreeze_sensor()
        elif cmd == "R" or cmd == "r":
            rval = Connector.restart_sensor()
        elif cmd == "S" or cmd == "s":
            rval = Connector.shutdown_sensor()
        elif cmd == "Q" or cmd == "q":
            Connector.logout()
            sys.exit(1)
        elif cmd == "P" or cmd == "p":
            return
        elif cmd == "?":
	    print "Sensor: " + sname + \
	          "\n\n" + do_sensor.__doc__
            continue
        else:
            print "Unknown command"
            continue

        check_message(rval)

def read_input_str(str):
    try:
        res = raw_input(str);
    except KeyboardInterrupt:
        sys.stdout.write("\n")
        return None
    else:
        return res

def do_logviewer (logviewer_func):
    round = 0
    while True:
        try:
            chunk = logviewer_func()
        except KeyboardInterrupt:
            Connector.stop_logread()
            break
        if chunk == None:
            print "Error reading log"
            break
        if round < 1:
            chunk = chunk[4:]
        round += 1
        sys.stdout.write(chunk)

def check_message(message, show=True):
    if message[:3] == "505":
        print "Request rejected"
        return 1
    if message[:3] == "509":
        print "Failed request"
        return 1
    if message[:3] not in opcodes_r:
        print "Protocol error"
        return 1
    else:
        if show:
            print message[4:]
    return 0

def save_file(fname, content):
    try:
        fp = open(fname, "w")
    except IOError, e:
        sys.stderr.write ("Failed to save file: " + e.strerror + "\n")
        return 1

    fp.write(content)
    fp.close()
    return 0

def open_file (fname):
    try:
        fp = open(fname, "r")
    except IOError, e:
        sys.stderr.write ("Failed to open file: " + e.strerror + "\n")
        return None
 
    content = fp.read()
    fp.close()
    return content
  
def verify_cb(conn, cert, errnum, depth, ok):
    if not ok:
        print "\n** Can't verify certificate ** Error " + `errnum` + " at depth " + `depth`\
              + "\n\n Issuer = %s\n Subject = %s\n MD5 Digest = %s\n" % \
              (cert.get_issuer(), cert.get_subject(),\
	       digest_cert(crypto.dump_certificate(crypto.FILETYPE_ASN1, cert)))
        if errnum == 18:    # Self signed certificate
            try:
                valid = raw_input("Do you accept certificate [y/n]? ")
            except KeyboardInterrupt:
                ok = 0
            if valid == 'y' or valid == 'Y':
                if store_certificate(crypto.dump_certificate(crypto.FILETYPE_PEM, cert)) == -1:
                    sys.stderr.write("Can't save certificate.\n")
                sys.stdout.write("\n"); ok = 1
            else:
                ok = 0
        if not ok:
            print "\nCertificate was rejected.\n"
    return ok

def read_configuration():
    try:
        fp = open(Confile).readlines()
    except IOError, e:
        fatalerror("Error: can't open " + Confile + ", " + e[1] + "\n")
    if stat.S_IMODE(os.lstat(Confile)[stat.ST_MODE]) & 077:
        fatalerror("Error: file " + Confile + " must have no more than -rwx------ (700) permissions.\n")
    else:
        c = len(fp) - 1
        while c >= 0:
            if fp[c] == "\n":
                del fp[c]
            c = c - 1
        try:
            uname = fp[0][:-1]
            passwd = fp[1][:-1]
            pindex = string.rfind(fp[2], ":")
            addr = fp[2][:pindex]
            port = fp[2][pindex + 1:-1]
        except:
            fatalerror("Error reading configuration file!\n")
    return addr, port, uname, passwd

def digest_cert(asn1cert):
    digest = md5.new(asn1cert).digest()
    return string.upper("%02x:" * len(digest) % tuple(map(ord, digest)))[:-1]

def store_certificate(cert):
    try:
        fp = open(CA_CERTS, "a")
        fp.write(cert)
    except IOError:
        return -1
    fp.close()

def fatalerror(error_message):
    sys.stderr.write(error_message)
    sys.exit(1)

if nmgrcomm.clientsocket.UseSSL < 1:
    fatalerror("Error: can't find pyOpenSSL module\n");


# Create connector
Connector = nmgrcomm.nodemanager_connector(CA_CERTS, verify_cb)

F_dict = {
            "--freeze": Connector.freeze_sensor,
            "--unfreeze": Connector.unfreeze_sensor,
            "--restart": Connector.restart_sensor,
            "--shutdown": Connector.shutdown_sensor,
            "--remove": Connector.remove_sensor,
            "--status": Connector.query_sensor_status,
            "--query": Connector.query_sensors,
            "--rulesets": Connector.query_rulesets,
            "--fetchrule": Connector.download_ruleset,
            "--sendrule": Connector.upload_ruleset,
            "--fetchconf": Connector.download_configuration,
            "--sendconf": Connector.upload_configuration,
            "--slog": Connector.read_sensor_log,
            "--nlog": Connector.read_nodemgr_log,
            "--sendlog": Connector.send_logged_data
         }

# --------------------------------------------------------------------
try:
    opts, args = getopt.getopt(sys.argv[1:], 'a:p:s:c:i',\
    ['help', 'version', 'freeze', 'unfreeze', 'restart', 'shutdown',\
     'remove', 'sendrule=', 'fetchrule=', 'sendconf', 'fetchconf',\
     'status', 'rulesets', 'slog', 'nlog', 'sendlog=', 'query'])
except:
    fatalerror("Error: try 'nuheclient.py --help' for more information.\n")
for o, a in opts:
    if o == '-i':
       Interactive = True
    elif o == '-a':
       x = string.rfind(a, ":")
       if x != -1:
           Addr = a[:x]
           Port = a[x+1:]
       else:
           Addr = a
    elif o == '-p':
        Port = a
    elif o == '-s':
       Sensor = a
    elif o == '-c':
       Confile = a

    elif o == '--version':
        sys.stdout.write("Nuhemanager cli client v. 0.01\n")
        sys.exit()
    elif o == '--help':
        print __doc__
        sys.exit()
    else:
        if Operation != None:
            fatalerror("Error: you can use only one operation at a time\n")

        if o == '--fetchrule' or \
           o == '--sendrule' or \
           o == '--sendlog':
            Userdata = a

        Operation = o
     

if Confile != None:
    Addr, Port, Uname, Passwd = read_configuration()

if Addr == None:
    fatalerror("Error: you must specify node manager address\n")

if type(Port) == type(''):
    try:
        Port = string.atoi(Port)
    except (ValueError, TypeError):
        fatalerror("Error: invalid port specification\n")

if not Interactive:
    if (Operation != "--query" and Sensor == None):
        if Operation != "--nlog":
            fatalerror("You must provide sensor name with -s command line option\n")

if Confile == None:
    Uname = get_name()
    Passwd = get_passwd()

# Main
main()
   
# Exit
sys.exit()
