/* event_handler.c
 *
 *  Copyright (C) 2006,2007 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "event_handler.h"

#define BUFSZ 4096

static struct list_s *search_node (struct list_s **elist, struct rule *rcell, char **event_id,
				   int size);
static int do_cmd_string (char *action, char **substr, char *cstring,
			  int substr_size);
static void do_action (struct event *ecell, char *sensorname);
static void run_external_cmd (struct event *ecell, char *cmd, char *sensorname);
static void free_event_data (struct list_s **head);
static int check_ignorelist (struct action *acell, char **event_id, int size,
			     int *index);
static int register_event (struct list_s **elist, struct rule *rcell, char *alertsrc,
			   char **event_id, int size);
static void update_event_status (struct list_s *ecell);
static int do_write_event_s (struct list_s **elist, char *eventfile);
static int do_read_event_s (struct list_s **elist, struct list_s *rlist, char *eventfile);

static pthread_mutex_t event_m = PTHREAD_MUTEX_INITIALIZER;

/* Searches elist for resetable events and resets them if allowed */
int
search_and_reset_event (struct internal *vars, char *entry)
{
  struct list_s *current;
  struct list_s **elist;
  struct event *payload;
  char *substr[SUBSTRMAX];
  int  flag;
  int i;

  flag = 0;
  elist = &vars->elist;

  pthread_mutex_lock (&event_m);
  current = *elist;

  while (current != NULL)
    {
      payload = (struct event *) current->payload;
      if (payload->acell->reset != NULL)
        {
          int rc;
          int idx;

          rc = -1;

          /* Loop through action struct reset d_array and perform Pcre match
           * for reset strings. NOTE: all events which share same action struct
           * (reset strings) will be reseted if substrings are matched
           */
          for (i = 0; i < d_array_get_size (payload->acell->reset); i++)
            {
              match_pcre (payload->acell->reset->element[i]->payload,
                          entry, substr, &rc);

              if (rc > 0)
                {
                  char *rstsub[SUBSTRMAX];
                  char **event_id;

                  event_id = substr;
                  /* substrtable is initialized */
                  if (payload->acell->substrtable[0] != -1)
                    {
                      int i;

                      /* Adjust reset string substrings to rule substrings */
                      for (i = 0; i < SUBSTRMAX; i++)
                        {
                          if (payload->acell->substrtable[i] != -1 &&
                              payload->acell->substrtable[i] < SUBSTRMAX)
                            rstsub[i] = substr[payload->acell->substrtable[i]];
                          else
                            rstsub[i] = NULL;
                        }
                      event_id = rstsub;
                    }

		  /* Compare event and matched reset string substrings */
	          for (idx = 0; idx < rc; idx++)
                    {
                      if (payload->acell->usesubstr[idx] == 1 &&
                          event_id[idx] != NULL)
                        {
                          if (strcmp (payload->event_id[idx], event_id[idx]) != 0)
                            {
                              flag = 0;
                              break;
                            }
                          else
                            flag = 1;
                        }
                    }

                  /* Free substrings */
                  for (idx = 0; idx < rc; idx++)
                    if (substr[idx] != NULL)
                      free (substr[idx]);

                  if (flag)
                    {
                      /* Found event: check if event is in phase where it can be reseted */
                      if (((struct do_action *) payload->acell->run_action->
                          element[payload->action_idx]->payload)->allowrst == 1)
                        payload->pending = -1;
                      break;
                    }
                }
            }
         }
       current = current->next;
    }
  pthread_mutex_unlock (&event_m);

  if (flag)
    return 0;
  return 1;
}

/* Match is found; we ignore event, register it or update it's status */
int
send_event (struct internal *vars, struct rule *rcell, char *alertsrc, 
            char **event_id, int size)
{
  struct list_s *current;
  struct list_s **elist;
  char logstr[BUFSZ];
  int index;

  elist = &vars->elist;

  /* First check that event_id has no same substrings that ignore list */
  if (check_ignorelist
      ((struct action *) rcell->aptr, event_id, size, &index))
    {
#if 0
      snprintf (logstr, BUFSZ,
		"[INFO] %s, rule #%d: action ignored based on criteria: %s",
		vars->sensorname, rcell->rulenr, event_id[index]);
      send_alarm ((const char *) logstr);
#endif
      return 1;
    }

  /* Check if event is registered */
  current = search_node (elist, rcell, event_id, size);
  if (current == NULL)
    {
      /* Register new event */
      pthread_mutex_lock (&event_m);
      if (register_event (elist, rcell, alertsrc, event_id, size) < 0)
	{
	  snprintf (logstr, BUFSZ,
		    "CRITICAL: registering new event failed, malloc() %s",
		    strerror (errno));
	  logwriter ((const char *) logstr);
          pthread_mutex_unlock (&event_m);

	  return -1;
	}
      pthread_mutex_unlock (&event_m);
    }
  else
    {
      /* Check if action is configured to use multiple alert
       * sources and if match is originating from same
       * source, before updating event
       */
      if (((struct event *) current->payload)->acell->multisource &&
           strcmp (alertsrc, ((struct event *)  current->payload)->alertsrc) != 0)
        return 1;

      /* Event exists, update it's status */
      update_event_status (current);
    }

  return 0;
}

static int
register_event (struct list_s **elist, struct rule *rcell, char *alertsrc,
                char **event_id, int size)
{
  struct list_s *new_node;
  struct event *payload;
  int i;

  /* Add new event */
  new_node = slist_push (elist);
  if (new_node == NULL)
    return -1;

  new_node->payload = (void *) malloc (sizeof (struct event));
  if (new_node->payload == NULL)
    return -1;

  /* Initialize event structure */
  payload = (struct event *) new_node->payload;
  payload->t_start = time (NULL);
  payload->acell = rcell->aptr;
  payload->rcell = rcell;
  payload->event_cnt = 1;
  payload->action_idx = 0;
  payload->id_size = size;
  payload->pending = 0;
  payload->ruledesc = rcell->desc;

  /* Event is indentified by matched substrings;
   * copy substrings to event structure
   */
  payload->event_id = (char **) malloc (size * sizeof (char *));
  if (payload->event_id == NULL)
    return -1;

  for (i = 0; i < size; i++)
    {
      payload->event_id[i] =
	(char *) calloc (strlen (event_id[i]) + 1, sizeof (char));
      if (payload->event_id[i] == NULL)
	return -1;
      strncpy (payload->event_id[i], event_id[i], strlen (event_id[i]));
    }

  payload->alertsrc = (char *) calloc (strlen (alertsrc) + 1, sizeof (char));
  if (payload->alertsrc == NULL)
    return -1;
  strncpy (payload->alertsrc, alertsrc, strlen (alertsrc));

  return 0;
}

/* Update counter of existing event */
static void
update_event_status (struct list_s *ecell)
{
  struct event *payload;

  pthread_mutex_lock (&event_m);
  payload = (struct event *) ecell->payload;
  payload->event_cnt++;
  pthread_mutex_unlock (&event_m);
}

/* mainloop for event handler */
void
monitor_events (struct internal *vars)
{
  struct list_s *current;
  struct list_s *previous;
  struct list_s **elist;
  struct event *payload;
  struct do_action *aptr;
  struct timeval timeout;
  time_t tm;
  int i;
  int aindex;
  int del_event;

  elist = &vars->elist;

  /* Load saved events from file to elist */
  pthread_mutex_lock (&event_m);
  if (do_read_event_s (elist, vars->rlist, vars->eventfile) != 0)
    logwriter("Loading of saved events failed");
  pthread_mutex_unlock (&event_m);

  while (1)
    {

#ifdef NODEMGR_SUPPORT
      /* FREEZE state means that only node manager connector thread should be
       * active; wait for state to change
       *
       * Current events are preserved as they are during FREEZE state
       */
      if (vars->run_stat == FREEZE)
        {
          pthread_mutex_lock (&freeze_sync.state_mutex);
          while (vars->run_stat == FREEZE)
            pthread_cond_wait (&freeze_sync.state_cv,
                               &freeze_sync.state_mutex);
          pthread_mutex_unlock (&freeze_sync.state_mutex);
        }
#endif

      /* quit_flag is set to exit state */
      if (quit_flag != RUNNING)
	{
	  pthread_mutex_lock (&event_m);
	  free_event_data (elist);
	  slist_delete_list (elist);
	  pthread_mutex_unlock (&event_m);
	  break;
	}

      /* Sleep for a while ... */
      timeout.tv_sec = 0;
      timeout.tv_usec = EVENT_SLEEPTIME;
      select (0, NULL, NULL, NULL, &timeout);

      pthread_mutex_lock (&event_m);
      current = *elist;
      previous = NULL;

      /* Loop through event list */
      while (current != NULL)
	{
	  del_event = 0;
	  payload = (struct event *) current->payload;
	  aindex = payload->action_idx;
	  aptr =
	    (struct do_action *) payload->acell->run_action->element[aindex]->
	    payload;

	  tm = time (NULL);

          /* Check if event is reseted and delete it */
          if (payload->pending == -1)
            {
              do_write_event_s (elist, vars->eventfile);
              del_event = 1;
            }

	  /* Only delete old actions that have positive event_max */
	  else if (aptr->event_max > 0 && payload->event_cnt <
	      aptr->event_max && (tm - payload->t_start) >= aptr->t_limit)
	    {
	      /* Time elapsed without maximum amount of events */
	      del_event = 1;
	    }

	  /* Run actions if they have reached event_cnt and time limit has exceeded, also
	   * actions with event_cnt 0 will be runned based on their time limit status
	   */
          else if (aptr->event_max > 0 && (payload->event_cnt > 0
                   && payload->event_cnt >= aptr->event_max
                   && (tm - payload->t_start) <= aptr->t_limit)
                   || (aptr->event_max == 0 && (tm - payload->t_start)
                   >= aptr->t_limit))
	     {
	       /* Trigger action for this event */
	       do_action (payload, vars->sensorname);

	       /* Check if we have subsequent actions */
	       if (payload->action_idx < d_array_get_size (payload->acell->run_action) - 1)
 		 {
		   if (((struct do_action *) payload->acell->run_action->
		        element[aindex + 1]->payload)->event_max > 0)
                     {
                       /* We're not interested in events; subsequent action will be
                        * runned after specified time
                        */
	               payload->event_cnt = 0;
                     }

		   payload->t_start = time (NULL);
		   payload->action_idx++;
                   /* Mark this event to be pending */
                   payload->pending = 1;
                   /* Save pending event structs to file */
                   do_write_event_s (elist, vars->eventfile);
		 }
	       else
		 {
		   /* No subsequent actions; delete event from list */
                   payload->pending = 0;
                   /* Save pending event structs to file */
                   do_write_event_s (elist, vars->eventfile);
		   del_event = 1;
		 }
	     }

          /* Delete event from list */
	  if (del_event)
	    {
	      /* Free event indentification strings */
	      for (i = 0; i < payload->id_size; i++)
		free (payload->event_id[i]);

              /* Free event_id pointer array and alertsrc */
              free (payload->event_id);
              free (payload->alertsrc);

              /* Remove node from event list */
              if (previous != NULL)
                previous->next = current->next;
              else
                *elist = current->next;

	      /* Free payload struct and event node */
              free (current->payload);
	      free (current);

	      break;
	    }
	  previous = current;
	  current = current->next;
	}
      pthread_mutex_unlock (&event_m);
    }
}

/* Write pending events to file */
static int
do_write_event_s (struct list_s **elist, char *eventfile)
{
  FILE *fp;
  struct list_s *current;
  struct event *payload;

  current = *elist;

  fp = fopen(eventfile, "w");
  if (fp == NULL)
    {
      char err[STRBUF];
      
      snprintf (err, STRBUF, "Can't open %s for writing, %s",
                eventfile, strerror(errno)); 
      logwriter(err);
      return 1;
    }
  
  while (current != NULL)
    {
      payload = (struct event *) current->payload;
      /* Found pending event */
      if (payload->pending == 1)
        {
          int i;

          /* Start indicator */
          fprintf (fp, "****\n");
         
          fprintf(fp, "%u\n", payload->id_size); 
          for (i = 0; i < payload->id_size; i++)
            fprintf(fp, "%s\n", payload->event_id[i]);

          fprintf(fp, "%s\n%u\n%u\n%u\n%s\n%ld\n", payload->alertsrc,
		  payload->event_cnt, payload->action_idx,
                  payload->pending, payload->ruledesc, 
                  (long) payload->t_start);

        }
      current = current->next;
    }

  fclose(fp);
  return 0;
}

/* Read saved events from file */
static int
do_read_event_s (struct list_s **elist, struct list_s *rlist, char *eventfile)
{
  FILE *fp;
  int i;
  char buffer[STRBUF];
  unsigned int rstrc;
  struct list_s *new_node;
  struct event *payload;

  payload = NULL;
  rstrc = 0;
  i = 0;

  fp = fopen(eventfile, "r");
  if (fp == NULL)
    {
      char err[STRBUF];

      snprintf (err, STRBUF, "Can't open %s for reading, %s",
                eventfile, strerror(errno));
      logwriter(err);
      return 1;
    }

  while (fgets (buffer, sizeof (buffer), fp) != NULL)
    {
       if (buffer[strlen (buffer) - 1] == '\n')
         buffer[strlen (buffer) - 1] = '\0';

      /* Found start of the event struct data */
      if (!strncmp (buffer, "****", 1))
        {
          rstrc = 1;
          i = 0;

          /* Create new event struct */
          payload = (struct event *) malloc (sizeof (struct event));
          if (payload == NULL)
            return -1;
          continue;
        }
      if (rstrc)
        {
          /* Read id_size */
          if (i == 0)
            {
              sscanf (buffer, "%u", &payload->id_size);
              payload->event_id = (char **) malloc (payload->id_size * sizeof (char *));
              if (payload->event_id == NULL)
                return -1;
            }
          else if (i <= payload->id_size)
            {
              payload->event_id[i-1] =
               (char *) calloc (strlen (buffer) + 1, sizeof (char));
              if (payload->event_id[i-1] == NULL)
                return -1;
              strncpy (payload->event_id[i-1], buffer, strlen (buffer));
            }
          else
           {
             /* First read alertsrc */
             if (i == payload->id_size + 1)
               {
                 payload->alertsrc = (char *) calloc (strlen (buffer) + 1, sizeof (char));
                 if (payload->alertsrc == NULL)
                   return -1;
                 strncpy (payload->alertsrc, buffer, strlen (buffer));
               }
             else if (i == payload->id_size + 2)
               sscanf (buffer, "%u", &payload->event_cnt);
             else if (i == payload->id_size + 3)
               sscanf (buffer, "%u", &payload->action_idx);
             else if (i == payload->id_size + 4)
               sscanf (buffer, "%u", &payload->pending);
#if 0
             else if (i == payload->id_size + 5)
               sscanf (buffer, "%ud", &payload->rulenr);
#endif
             else if (i == payload->id_size + 5)
               {
                 payload->ruledesc = (char *) calloc (strlen (buffer) + 1, sizeof (char));
                 if (payload->ruledesc == NULL)
                   return -1;
                 strncpy (payload->ruledesc, buffer, strlen (buffer));
               }
             else if (i == payload->id_size + 6)
               {
                 struct list_s *current;
                 long tmp;
                 int out;

                 out = 0;
                 sscanf (buffer, "%ld", &tmp);
                 payload->t_start = (time_t) tmp;
                 
                 /* We are now ready; find corresponding rule structure */
                 current = rlist;
                 while (current != NULL)
                   {
                     if (!strcmp(((struct rule *) current->payload)->desc,
                                  payload->ruledesc))
                       {
                         payload->rcell = (struct rule *) current->payload;
                         payload->acell = ((struct rule * )
                                           (struct action *) current->payload)->aptr;
                         free (payload->ruledesc);
                         payload->ruledesc = ((struct rule *)current->payload)->desc;

                         /* Add new event to elist */
                         new_node = slist_push (elist);
                         if (new_node != NULL)
                           {
                             new_node->payload = (void *) payload;
                             out = 1;
                           }
                         break;
                       }
                     current = current->next;
                   }
                 /* No corresponding rule found */
                 if (!out)
                   {
                     int i;

                     logwriter ("Warning: unable to initialize loaded event, skipping");
                     for (i = 0; i < payload->id_size; i++)
                       free (payload->event_id[i]);

                     free (payload->alertsrc);
                     free (payload);
                   }
                 rstrc = 0;
               }
           }
          i++;
        }
    }
  fclose (fp);
  return 0;
}

/* Wrapper for do_cmd_string and run_external functions */
static void
do_action (struct event *ecell, char *sensorname)
{
  char cstring[BUFSZ] = { 0 };
  char logstr[BUFSZ];

  /* If logthreshold is set to EVENT, we log only when event action is
   * triggered
   */

  /* Generate alarms here */
  if (ecell->acell->ltvalue == EVENT)
    {
      char *rstr;
      char alarmstr[LOGBUF];

      if (ecell->action_idx == 0)
        {
          rstr =
             generate_alarm (ecell->rcell, ecell->alertsrc,
                             sensorname, ecell->event_id[0]);
     
          if (rstr != NULL)
            {
              snprintf (alarmstr, LOGBUF,
                        "%s\nAlert repeated %d times", rstr, ecell->event_cnt);
              send_alarm ((const char *) alarmstr);
              free (rstr);
            }
        }
    }

  if (do_cmd_string
      (((struct do_action *) ecell->acell->run_action->
	element[ecell->action_idx]->payload)->action, ecell->event_id,
       cstring, ecell->id_size) == -1)
    {
      snprintf (logstr, BUFSZ,
		"[FAILURE] %s, error: indexing substring out of bounds, can't execute '%s' command for rule #%d",
		sensorname, ((struct do_action *) ecell->acell->run_action->
		 element[ecell->action_idx]->payload)->action,
		ecell->rcell->rulenr);

      send_alarm ((const char *) logstr);
    }
  else
    run_external_cmd (ecell, cstring, sensorname);
}

/* This function copies action string with modifications to cstring
 * ($0-9 characters are substituded with corresponding Pcre substrings)
 */
static int
do_cmd_string (char *action, char **substr, char *cstring, int substr_size)
{
  int i = 0;
  int x = 0;
  int y;
  int sidx;
     
  while (i < strlen (action))
    {
      if (i >= BUFSZ)
	return 0;
      if (action[i] == '$' && i < strlen (action) - 1)
	{
	  sidx = atoi (&action[++i]);
	  if (sidx > substr_size - 1)
	    return -1;
	  for (y = 0; y < strlen (substr[sidx]); y++)
	    cstring[x++] = substr[sidx][y];
	}
      else
	cstring[x++] = action[i];
      i++;
    }
  return 0;
}

static void
run_external_cmd (struct event *ecell, char *cmd, char *sensorname)
{
  FILE *fp;
  int rval;
  char logstr[BUFSZ];

  /* Check what kind of action indentifier we have */
  if (ecell->acell->id == CMD)
    {
      /* Run external command specified by rule action */

      /* FIX this to not use popen() */
      fp = popen (cmd, "r");
      if (fp == NULL)
	{
	  snprintf (logstr, BUFSZ,
		    "[FAILURE] %s, error: can't execute '%s' for rule #%d, %s", 
                    sensorname, cmd,
		    ecell->rcell->rulenr, strerror (errno));

	  send_alarm ((const char *) logstr);
	}
      else
	{
	  rval = pclose (fp);

	  snprintf (logstr, BUFSZ,
		    "[ACTION] %s, executed command '%s' for rule #%d with return value of %d",
		    sensorname, cmd, ecell->rcell->rulenr, rval);

	  send_alarm ((const char *) logstr);
	}
    }
#if 0
  else if (ecell->acell->id == SHELL)
    {
      /* TODO */
    }
#endif
}

/* This function validates if event should be ignored (based on ignorelist) */
static int
check_ignorelist (struct action *acell, char **event_id, int size, int *index)
{
  int i;
  int x;

  char *ign_data;

  for (i = 0; i < d_array_get_size (acell->ignore); i++)
    {
      for (x = 0; x < size; x++)
	{
          ign_data = (char *) acell->ignore->element[i]->payload;
	  if (strcmp (ign_data, event_id[x]) == 0)
	    {
	      *index = x;
	      return 1;
	    }
	}
    }
  return 0;
}

/* Searches event list, returns pointer to event structure if found
 * or NULL otherwise
 */
static struct list_s *
search_node (struct list_s **elist, struct rule *rcell, char **event_id, int size)
{
  struct list_s *current;
  struct event *payload;
  int i;
  int flag;

  flag = 0;
  pthread_mutex_lock (&event_m);
  current = *elist;

  while (current != NULL)
    {
      payload = (struct event *) current->payload;

      /* Check that we have correct rule */
      if (payload->rcell->rulenr == rcell->rulenr)
        {
          for (i = 0; i < size; i++)
	    {
	      if (payload->acell->usesubstr[i] == 1 &&
                event_id[i] != NULL)
	        {
                  /* Check that we have correct event */
	          if (strcmp (payload->event_id[i], event_id[i]) != 0)
	            {
		      flag = 0;
		      break;
		    }
	          else
                    flag = 1;
	        }
	    }

          if (flag == 1)
	    {
	      /* Event found */
	      pthread_mutex_unlock (&event_m);
	      return current;
	    }
        }

      current = current->next;
    }
  pthread_mutex_unlock (&event_m);

  return NULL;
}

static void
free_event_data (struct list_s **head)
{
  struct list_s *current;
  struct list_s *next;
  struct event *cell;
  int i;

  current = *head;
  
  while (current != NULL)
    {
      next = current->next;
      cell = (struct event *) current->payload;

      for (i = 0; i < cell->id_size; i++)
	free (cell->event_id[i]);

      free (cell->event_id);
      free (cell->alertsrc);
      
      current = next;
    }
}
