/* nuheclient.c
 *
 * Copyright (C) 2007 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "nuheclient.h"

/* Timeout in seconds between connection retry */
#define CONN_RETRY_TIMEOUT 5

/* 0 means infinite retrying if connector fails */
#define CONN_RETRIES 0

static int open_connection (struct internal *vars);
static int handle_connection (SSL * ssl, struct internal *vars);
static enum msgtype get_message_opcode (char *msg, struct internal *vars);
static char *read_sensor_file (char *fpath, enum msgtype type);
static void get_message_payload (char *srcbuf, char *dstbuf, unsigned int dstlen);
static int handle_file_operation (SSL * ssl, char *ofile, enum msgtype type);
static int send_pending_messages (SSL * ssl);
static int save_sensor_file (char *cfile, char *fcontents);

/* Mutex for msglist */
static pthread_mutex_t msglist_m = PTHREAD_MUTEX_INITIALIZER;

/* list_s for alert messages */
static struct list_s *msglist;

/* Holds original status of Nuhe */
static enum status orig_state;

/* Protocol phases */
enum cphase
{
  STARTCOM = 0,
  CONCHECK
};

/* Protocol message types */
enum msgtype
{
  CLNTHELLO = 0,
  SRVRHELLO,
  SRVRPING,
  SRVRREQRS,
  SRVRGETRN,
  SRVRSENDRN,
  SRVRREQCN,
  SRVRSENDCN,
  SRVRREQALRT,
  SRVRREQBOOT,
  SRVRREQQUIT,
  SRVRREQFRZ,
  SRVRREQUNFRZ,
  REQRJCT,
  UNKNOWN
};

void
nuhe_connector (struct internal *vars)
{
  struct timeval timeout;
  int ci;

  orig_state = vars->run_stat;
  ci = 0;

  /* Initialize mutex & cv for FREEZE state */
  pthread_mutex_init (&freeze_sync.state_mutex, NULL);
  pthread_cond_init (&freeze_sync.state_cv, NULL);

  if (init_openssl () > 0)
    logwriter
      ("Error: OpenSSL initialization failed, node manager connector can't be started!");

  else
    {
      logwriter ("Nuhe node manager connector started");
      while (1)
	{
	  if (open_connection (vars) < 0)
	    {
	      /* bail out ? */
	      if (CONN_RETRIES > 0 && ci >= CONN_RETRIES)
		{
		  logwriter
		    ("Error: connection to node manager failed, giving up");
                  break;
		}
	      /* Something went wrong
	       * wait for CONN_RETRY_TIMEOUT and try again ...
	       */
	      timeout.tv_sec = CONN_RETRY_TIMEOUT;
	      timeout.tv_usec = 0;

	      select (0, NULL, NULL, NULL, &timeout);
	      ci++;
	    }

	  /* Check if Nuhe has received SIGTERM, SIGINT or
	   * SIGHUP signal
	   */
	  if (quit_flag != RUNNING)
	    break;
	}
    }

  pthread_mutex_destroy (&freeze_sync.state_mutex);
  pthread_cond_destroy (&freeze_sync.state_cv);
}

static int
open_connection (struct internal *vars)
{
  SSL_CTX *ctx;
  SSL *ssl;

  long err;
  int closedc;
  int ret;

  char hconn[STRBUF] = { 0 };

  closedc = 0;
  snprintf (hconn, sizeof (hconn), "%s:%d", vars->manageraddr,
	    vars->managerport);

  seed_prng (1024);

  /* Create SSL context, CA directory is currently NULL */
  ctx = create_ssl_context (vars->certificate, vars->privatekey, vars->cafile, NULL);
  if (ctx == NULL)
    return -1;

  /* Create SSL connection to node manager */
  ssl = create_ssl_connection (ctx, hconn);
  if (ssl == NULL)
    return -1;

  ret = SSL_connect (ssl);
  if (ret <= 0)
    {
      int ec;
      char estr[STRBUF] = { 0 };

      ec = SSL_get_error (ssl, ret);
      if (ec == SSL_ERROR_SYSCALL)
	snprintf (estr, sizeof (estr),
		  "Error: TLS/SSL handshake was not successful, %s",
		  strerror (errno));
      else
        {
          char err[128] = { 0 };
          ERR_error_string (ERR_get_error(), err);
	  snprintf (estr, sizeof (estr),
	       	    "Error: TLS/SSL handshake was not successful, %s",
                     err);
        }

      logwriter (estr);
      closedc = 1;
      goto out;
    }

  /* Post connection validation */
  if ((err = post_conn_check (ssl, vars->manageraddr)) != X509_V_OK)
    {
      char data[STRBUF] = { 0 };

      snprintf(data, sizeof (data), "Error: peer certificate, %s",
               X509_verify_cert_error_string (err));
      logwriter (data);
      closedc = 1;
      goto out;
    }

  /* Handle our connection */
  if (handle_connection (ssl, vars) < 0)
    {
      /* Peer probably closed socket */
      logwriter
	("Error: node manager shutted connection, retrying to connect");
      closedc = 1;
    }

out:
  SSL_shutdown (ssl);
  SSL_free (ssl);
  SSL_CTX_free (ctx);

  if (closedc > 0)
    return -1;

  return 0;
}

static int
handle_connection (SSL * ssl, struct internal *vars)
{
  char txbuf[BUFSIZE] = { 0 };
  char managername[STRBUF] = { 0 };
  char *rxbuf;

  unsigned int rval;
  unsigned int darr_sz;
  unsigned int i;
  int err;
  int sock;

  enum cphase task;
  struct rulefiles *rstrc;
  struct timeval timeout;
  fd_set fds;

  /* Initial protocol phase */
  task = STARTCOM;
  /* Read channel fd for ssl */
  sock = SSL_get_rfd (ssl);

  while (1)
    {
      /* quit_flag is set to exit state */
      if (quit_flag != RUNNING)
	{
          /* Make sure that FREEZE state threads are signaled; they
           * should be able to perform clean-up tasks before exit
           */
          pthread_mutex_lock (&freeze_sync.state_mutex);
          vars->run_stat = orig_state;
          pthread_cond_broadcast (&freeze_sync.state_cv);
          pthread_mutex_unlock (&freeze_sync.state_mutex);

          if (quit_flag == QUIT)
            write_sock (ssl, "391 BYE\r\n");
          else if (quit_flag == SHUTDOWN)
            write_sock (ssl, "394 BYE\r\n");
          else if (quit_flag == RESTART)
            write_sock (ssl, "202 RESTARTING\r\n");
	  return 0;
	}

      if (task == STARTCOM)
	{
	  /* Let's say hello to server */
	  if (vars->run_stat != DEFUNC)
	    snprintf (txbuf, sizeof (txbuf), "151 %s HELLO\r\n",
		      vars->sensorname);
	  else
	    /* DEFUNC state hello */
            snprintf (txbuf, sizeof (txbuf), "151 %s DEFUNC\r\n",
	              vars->sensorname);

	  if (write_sock (ssl, txbuf) < 1)
	    {
              enum msgtype opcode;

	      rval = read_sock (ssl, &rxbuf);
	      if (rval != 0)
		return -1;

	      opcode = get_message_opcode (rxbuf, vars);
	      if (opcode == SRVRHELLO)
		{
                  char errmsg[LOGBUF];

		  /* Fetch node manager name */
		  get_message_payload (rxbuf, managername, STRBUF);
		  if (managername[strlen (managername) - 1] == '\n')
		    {
		      managername[strlen (managername) - 1] = '\0';
		    }

		  logwriter ("Connection to node manager established");

                  /* Report errors and warnings to node manager 
                   ********************************************
                   */
 
                  /* We are running in DEFUNC state; report error to node manager*/
                  if (vars->run_stat == DEFUNC)
                    {
                      snprintf (errmsg, LOGBUF, "[ERROR] %s, %s",
                              vars->sensorname, vars->errstr);
                      send_alarm ((const char *) errmsg);
                    }

                  /* Check if we are running with .backup configuration file 
                   * and report it to node manager
                   */
                  if (strstr (vars->cfile, ".backup") != NULL)
                    {
                      snprintf (errmsg, LOGBUF,
                                "[WARNING] %s, running with configuration file backup",
                                vars->sensorname);
                      send_alarm ((const char *) errmsg);
                    }

                  /* We're running with file monitor disabled; report it no node manager */
                  if (vars->active_fmon < 1)
                    {
                      snprintf (errmsg, LOGBUF,
                                "[WARNING] %s, running with file monitor disabled",
                                vars->sensorname);
                      send_alarm ((const char *) errmsg);
                    }

                  /* Handshake is done, move to CONCHECK mode */
                  task = CONCHECK;
		}
              free (rxbuf);
	    }
	  else
	    return -1;
	}
      else if (task == CONCHECK)
	{
	  /* This is so called idle mode; If select timeouts we
           * send pending alert messages for node manager, otherwise
           * we listen ALIVE messages from node manager, reply to them
           * and switch between tasks
	   */

          enum msgtype task_n;

	  FD_ZERO (&fds);
	  FD_SET (sock, &fds);

	  timeout.tv_sec = 1;
	  timeout.tv_usec = 0;

	  err = select (sock + 1, &fds, NULL, NULL, &timeout);

	  /* select timeout; send pending alarms to node manager */
	  if (err == 0)
	    {
	      if (send_pending_messages (ssl) < 0)
                return -1;
	      continue;
	    }

	  /* select error */
	  if (err == -1)
	    continue;

	  /* We are ready to read */
	  rval = read_sock (ssl, &rxbuf);
	  if (rval != 0)
	    return -1;

	  task_n = get_message_opcode (rxbuf, vars);

          /* free rxbuf for all other tasks */
          if (task_n == SRVRGETRN || task_n == SRVRREQALRT ||
              task_n == SRVRSENDCN || task_n == SRVRSENDRN)
            { }
          else
            {
              if (rxbuf != NULL)
                free (rxbuf);
            }

	  if (task_n == SRVRPING)
	    {
	      /* Send ALIVE reply to server */
	      if (write_sock (ssl, "154 OK\r\n") < 1)
		continue;
	      else
		return -1;
	    }
          else if (task_n == SRVRREQQUIT)
            {
              logwriter("Node manager requested shutdown, exiting...");
              quit_flag = SHUTDOWN;
              continue;
            }
          else if (task_n == SRVRREQBOOT)
            {
              logwriter("Node manager requested restart, restarting...");
              quit_flag = RESTART;
              continue;
            }
          else if (task_n == SRVRREQFRZ && vars->run_stat != FREEZE)
            {
              vars->run_stat = FREEZE;
              logwriter("Node manager requested freezing, sensor stopped");
              if (write_sock (ssl, "212 FREEZED\r\n") > 0)
                return -1;
              continue;
            }
          else if (task_n == SRVRREQUNFRZ)
            {
              /* Change state back to RUNNING and signal waiting threads */
	      pthread_mutex_lock (&freeze_sync.state_mutex);
              vars->run_stat = orig_state;
	      pthread_cond_broadcast (&freeze_sync.state_cv);
	      pthread_mutex_unlock (&freeze_sync.state_mutex);
           
              logwriter("Node manager requested unfreezing, sensor running");
              if (write_sock (ssl, "214 UNFREEZED\r\n") > 0)
                return -1;
              continue; 
            }
	  else if (task_n == SRVRREQRS)
	    {
	      memset (txbuf, 0, sizeof (txbuf));
	      snprintf (txbuf, sizeof (txbuf), "162 ");
	      darr_sz = d_array_get_size (vars->rules);
	      for (i = 0; i < darr_sz; i++)
		{
		  rstrc =
		    (struct rulefiles *) vars->rules->element[i]->payload;
		  strncat (txbuf, rstrc->fname, sizeof (txbuf));
		  if (i + 1 < darr_sz)
		    strncat (txbuf, " ", sizeof (txbuf));
		}
	      strncat (txbuf, CRLF, sizeof (txbuf));

	      /* Send reply to node manager */
	      if (write_sock (ssl, txbuf) < 1)
		task = CONCHECK;
	      else
		return -1;
	    }
	  else if (task_n == SRVRGETRN)
	    {  
              int out;
	      char rulefile[STRBUF] = { 0 };

              out = 0;
	      get_message_payload (rxbuf, rulefile, STRBUF);
              free (rxbuf);

	      darr_sz = d_array_get_size (vars->rules);
	      for (i = 0; i < darr_sz; i++)
		{
		  rstrc =
		    (struct rulefiles *) vars->rules->element[i]->payload;
		  if (!strcmp (rulefile, rstrc->fname))
		    {
		      if (handle_file_operation (ssl, rstrc->fpath, task_n) <
			  0)
			return -1;
                      out = 1;
		      break;
		    }
		}
              if (!out)
                {
                  if (write_sock (ssl, "509 REQUEST FAILED\r\n") > 0)
                    return -1;
                }
	      task = CONCHECK;
	    }
	  else if (task_n == SRVRREQCN)
	    {
	      if (handle_file_operation (ssl, vars->cfile, task_n) < 0)
		return -1;
	      task = CONCHECK;
	    }
          else if (task_n == SRVRSENDCN || task_n == SRVRSENDRN)
            {
              char *fname;
              char *fcontents;
              char *payload;
              int rval;

              payload = (char *) calloc (strlen (rxbuf) + 1, sizeof (char));
              if (payload == NULL)
                return -1;

              get_message_payload (rxbuf, payload, strlen (rxbuf));
              free (rxbuf);

              /* If task_n is SRVRSENDRN, we must extract
               * ruleset name from message
               */
              if (task_n == SRVRSENDRN)
                {
                  char *rnptr;
                  char fpath[STRBUF] = { 0 };

                  /* adjust rnptr to message payload */
                  rnptr = &payload[0];
                  while (!isspace ((int) *rnptr))
                    rnptr++;
                  *rnptr = '\0';

                  snprintf (fpath, sizeof (fpath), "%s/%s", vars->ruledir, payload);
                      
                  fname = &fpath[0];
                  fcontents = ++rnptr;
                }
              else
                {
                  fname = vars->cfile;
                  fcontents = payload;
                }

              /* Save configuration/ruleset file */
              rval = save_sensor_file (fname, fcontents);
              free (payload);

              if (rval != 0)
                {
                  if (write_sock (ssl, "509 REQUEST FAILED\r\n") > 0)
                    return -1;
                }
              else
                {
                  if (write_sock (ssl, (task_n == SRVRSENDCN ? 
                                  "174 OK\r\n" : "166 OK\r\n")) > 0)
                    return -1;
                }
              task = CONCHECK;
            }
	  else if (task_n == SRVRREQALRT)
	    {
              char *alertstr;

	      if (vars->manageralrt > 0)
		{
                  alertstr = (char *) calloc (strlen (rxbuf) + 1, sizeof (char));
                  if (alertstr == NULL)
                    {
                      free (rxbuf);
                      logwriter("CRITICAL: allocating alarm event failed, calloc()");
                      return 1;
                    }

		  /* Get payload from message */
		  get_message_payload (rxbuf, alertstr, BUFSIZE);
                  free (rxbuf);

		  /* Process alert message */
		  process_alert_data (vars, managername, alertstr);
                  free (alertstr);

		  if (write_sock (ssl, "182 OK\r\n") > 0)
		    return -1;
		  task = CONCHECK;
		}
	      else
		{
                  free (rxbuf);
		  if (write_sock (ssl, "505 REQUEST REJECTED\r\n") > 0)
		    return -1;
		  task = CONCHECK;
		}
	    }
	  else if (task_n == UNKNOWN)
	    {
	      /* Received something weird, send error message and
	       * switch to STARTCOM state
	       */
	      if (write_sock (ssl, "507 UNKNOWN MESSAGE\r\n") > 0)
		return -1;
	      task = STARTCOM;
	    }
          else if (task_n == REQRJCT)
            {
              /* Reject node manager request */
              if (write_sock (ssl, "505 REQUEST REJECTED\r\n") > 0)
                return -1;
              task = CONCHECK;
            }
	}
    }
}

/* Extracts message payload by removing protocol opcode and delimiter */
static void
get_message_payload (char *srcbuf, char *dstbuf, unsigned int dstlen)
{
  char *cptr;
  char *fptr;
  unsigned int i;

  i = 0;
  fptr = &dstbuf[0];

  /* adjust fptr to message payload */
  cptr = &srcbuf[0];
  while (!isspace ((int) *cptr))
    cptr++;
  cptr++;

  while (isgraph ((int) *cptr) || isspace ((int) *cptr))
    {
      if (*cptr == '\r' && *(cptr + 1) == '\n')
        break;
      if (i >= dstlen)
        break;
      else
        *fptr++ = *cptr++;
      i++;
    }
}

static int
save_sensor_file (char *fname, char *fcontents)
{
  FILE *fp;
  char *buptr;

  buptr = strstr (fname, ".backup"); 
  if (buptr == NULL)
    {
      /* Take backup from working configuration file */

      char *bufile;

      bufile = (char *) calloc (strlen (fname) + 10, sizeof (char));
      if (bufile == NULL)
        return -1;

      sprintf (bufile, "%s.backup", fname);
      rename (fname, bufile);
      
      free (bufile);
    }
  else
    {
      /* We do not want to back up broken configuration
       * file because we are already running with working
       * .backup configuration file
       */ 
      *buptr = '\0';
    }

  fp = fopen (fname, "w");
  if (fp == NULL)
    return 1;

  if (write (fileno (fp), fcontents, strlen (fcontents)) < 0)
    return 1;
  if (fsync (fileno (fp)) < 0)
    return 1;

  fclose (fp);

  return 0;
}

static int
handle_file_operation (SSL * ssl, char *ofile, enum msgtype type)
{
  int wval;
  char *fmsg;

  fmsg = read_sensor_file (ofile, type);
  if (fmsg != NULL)
    {
      /* Message was constructed by read_sensor_file, send
       * it now to node manager
       */
      wval = write_sock (ssl, fmsg);
      free (fmsg);
      if (wval != 0)
	return -1;
    }
  else
    {
      /* Ouch ... */
      if (write_sock (ssl, "509 REQUEST FAILED\r\n") > 0)
	return -1;
    }
  return 0;
}

static char *
read_sensor_file (char *fpath, enum msgtype type)
{
  FILE *fp;
  char buffer[512] = {
    0
  };
  char *fmsg;
  char *newp;
  unsigned int r;

  r = 1;

  fp = fopen (fpath, "r");

  if (fp == NULL)
    return NULL;

  fmsg = (char *) calloc (BUFSIZE, sizeof (char));
  if (fmsg == NULL)
    return NULL;

  /* Prefix message with protocol opcode */
  if (type == SRVRGETRN)
    snprintf (fmsg, BUFSIZE, "164 ");
  else if (type == SRVRREQCN)
    snprintf (fmsg, BUFSIZE, "172 ");

  while (fgets (buffer, sizeof (buffer), fp) != NULL)
    {
      if (r > 1)
	{
	  newp = (char *) calloc  (r * BUFSIZE, sizeof (char));
	  if (newp == NULL)
	    return NULL;

          strcpy (newp, fmsg);
          free (fmsg);
	  fmsg = newp;
	}
      strncat (fmsg, buffer, BUFSIZE);
      r++;
    }
  fclose (fp);

  /* Terminate message with CRLF delimiter */
  strncat (fmsg, CRLF, r * BUFSIZE);

  return fmsg;
}

/* Check message opcode */
static enum msgtype
get_message_opcode (char *msg, struct internal *vars)
{
  /* Do not process stubb messages */
  if (strlen (msg) < 3)
    return UNKNOWN;

  /* If we are in FREEZE state only unfreezing is allowed */
  if (vars->run_stat == FREEZE)
    {
      if (!strncmp (msg, "213", 3))
        /* We're free to go! */
        return SRVRREQUNFRZ;

      return REQRJCT;
    }

  /* General messages for RUNNING and DEFUNC state
   ***********************************************
   */ 
  if (!strncmp (msg, "152", 3))
    /* Node manager has replied to our hello message */
    return SRVRHELLO;
  else if (!strncmp (msg, "153", 3))
    /* Node manager sends heartbeat for us */
    return SRVRPING;
  else if (!strncmp (msg, "161", 3))
    /* Node manager has requested our rulesets */
    return SRVRREQRS;
  else if (!strncmp (msg, "163", 3))
    /* Send specific ruleset for node manager */
    return SRVRGETRN;
  else if (!strncmp (msg, "165", 3))
    /* Node manager uploads specific ruleset for us */
    return SRVRSENDRN;
  else if (!strncmp (msg, "171", 3))
    /* Node manager requested our configuration file */
    return SRVRREQCN;
  else if (!strncmp (msg, "173", 3))
    /* Node manager uploads configuration file for us */
    return SRVRSENDCN;
  else if (!strncmp (msg, "201", 3))
    /* Node manager request restart */
    return SRVRREQBOOT;
  else if (!strncmp (msg, "393", 3))
    /* Node manager disconnects us */
    return SRVRREQQUIT;

  /* DEFUNC state does not support all messages */
  if (vars->run_stat == DEFUNC)
    return REQRJCT;
  else
    {
      if (!strncmp (msg, "181", 3))
        /* Node manager has sent alert request */
        return SRVRREQALRT;
      else if (!strncmp (msg, "211", 3))
        /* Node manager freezes us*/
        return SRVRREQFRZ;

      return UNKNOWN;
    }
}

/* Check if we have pending alert messages and send them 
 * to node manager
 */
static int
send_pending_messages (SSL * ssl)
{
  char *alarmstr;
  char *wiremsg;
  struct list_s *current;

  pthread_mutex_lock (&msglist_m);
  current = msglist;
  while (current != NULL)
    {
      alarmstr = (char *) current->payload;
      wiremsg = (char *) calloc (strlen (alarmstr) + 10, sizeof (char));
      if (wiremsg == NULL)
        {
          logwriter("CRITICAL: allocating alarm message for node manager failed, calloc()");
          return 1;
        }

      snprintf (wiremsg, strlen (alarmstr) + 10, "191 %s\r\n", alarmstr);

      if (write_sock (ssl, wiremsg) > 0)
	{
	  free (wiremsg);
	  pthread_mutex_unlock (&msglist_m);
	  return -1;
	}
      free (wiremsg);
      current = current->next;
    }

  /* Reset msglist */
  slist_delete_list (&msglist);
  pthread_mutex_unlock (&msglist_m);

  return 0;
}

/* Add alerts received from Nuhe sensor to msglist and
 * process them later in CONCHECK state
 */
int
send_message_nodemgr (const char *msg)
{
  struct list_s *new_node = NULL;

  pthread_mutex_lock (&msglist_m);
  new_node = slist_append_node (&msglist);
  if (new_node == NULL)
    {
      pthread_mutex_unlock (&msglist_m);
      return -1;
    }

  new_node->payload = (char *) calloc (strlen (msg) + 1, sizeof (char *));
  if (new_node->payload == NULL)
    {
      pthread_mutex_unlock (&msglist_m);
      return -1;
    }

  strncpy (new_node->payload, msg, strlen (msg));
  pthread_mutex_unlock (&msglist_m);

  return 0;
}
