/* nuhesock.c
 *
 * Copyright (C) 2007 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include "nuhesock.h"

void
handle_error (const char *file, int fileno, const char *msg)
{
  char buf[256] = { 0 };
  char err[128];

  ERR_error_string (ERR_get_error (), err);

  snprintf (buf, sizeof (buf), "%s, %s", msg, err);
  logwriter (buf);
}

int
init_openssl (void)
{
  if (!SSL_library_init ())
    return 1;

  SSL_load_error_strings ();
  return 0;
}

int
seed_prng (int bytes)
{
  if (!RAND_load_file ("/dev/urandom", bytes))
    return 0;
  return 1;
}

SSL_CTX *
create_ssl_context (char *certfile, char *pkeyfile, char *cafile, char *cadir)
{
  SSL_CTX *ctx;
  int err;

  ctx = SSL_CTX_new (SSLv23_method ());
  SSL_CTX_set_mode (ctx, SSL_MODE_AUTO_RETRY);

  err = 0;
  if (SSL_CTX_load_verify_locations (ctx, cafile, cadir) != 1)
    {
      int_error ("Error: can't load CA file");
      err = 1;
    }
  if (SSL_CTX_set_cipher_list (ctx, CIPHERS) != 1)
    {
      int_error ("Error: no valid ciphers");
      err = 1;
    }
  if (SSL_CTX_use_certificate_chain_file (ctx, certfile) != 1)
    {
      int_error ("Error: can't load certificate chain");
      err = 1;
    }
  if (SSL_CTX_use_PrivateKey_file (ctx, pkeyfile, SSL_FILETYPE_PEM) != 1)
    {
      int_error ("Error: can't load private key");
      err = 1;
    }

  if (err)
    {
      free (ctx);
      return NULL;
    }

  SSL_CTX_set_options (ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);
  SSL_CTX_set_verify (ctx, SSL_VERIFY_PEER, verify_cb);
  SSL_CTX_set_verify_depth (ctx, 4);

  return ctx;
}

SSL *
create_ssl_connection (SSL_CTX * ctx, char * hconn)
{
  BIO *conn;
  SSL *ssl;

  /* Open connection */
  conn = BIO_new_connect (hconn);
  if (!conn)
    {
      logwriter ("Error: creating connection to node manager");
      return NULL;
    }

  if (BIO_do_connect (conn) <= 0)
    {
      logwriter ("Error: connecting to node manager");
      return NULL;
    }

  if (!(ssl = SSL_new (ctx)))
    {
      logwriter ("Error: creating an SSL context");
      return NULL;
    }

  SSL_set_bio (ssl, conn, conn);

  return ssl;
}

int
verify_cb (int ok, X509_STORE_CTX * store)
{

  if (!ok)
    {
      char *ename;
      char *esubj;
      char errmsg[1024]
                  = { 0 };
      int depth;
      int err;
      X509 *cert;


      cert = X509_STORE_CTX_get_current_cert (store);
      depth = X509_STORE_CTX_get_error_depth (store);
      err = X509_STORE_CTX_get_error (store);
      ename = X509_NAME_oneline (X509_get_issuer_name (cert), NULL, 256);
      esubj = X509_NAME_oneline (X509_get_subject_name (cert), NULL, 256);

      snprintf (errmsg, sizeof (errmsg),
		"Error with certificate at depth: %i\n issuer  = %s\n subject = %s\n err %i:%s",
		depth, ename, esubj, err,
		X509_verify_cert_error_string (err));

      free (ename);
      free (esubj);

      logwriter (errmsg);
    }
  return ok;
}

long
post_conn_check (SSL * ssl, char *host)
{
  X509 *cert;
  X509_NAME *subj;
  char data[256] = { 0 };


  cert = SSL_get_peer_certificate (ssl);
  subj = X509_get_subject_name (cert);

  if (!cert || !host)
    goto err_occured;

  /* commonName not found */
  if (X509_NAME_get_text_by_NID (subj, NID_commonName, data, 256) < 0)
    goto err_occured;
  if (strcasecmp (data, host) != 0)
    goto err_occured;

  X509_free (cert);

  return SSL_get_verify_result (ssl);

err_occured:
  if (cert)
    X509_free (cert);
  return X509_V_ERR_APPLICATION_VERIFICATION;
}

#if 0
int
read_sock (SSL * ssl, char *buf)
{
  int err;
  int nread;

  for (nread = 0; nread < DBUFSZ; nread += err)
    {
      err = SSL_read (ssl, buf + nread, DBUFSZ - nread);
      /* Check if we have reached end of message */
      if (buf[strlen (buf) - 2] == '\r' && buf[strlen (buf) - 1] == '\n')
	break;
      /* Connection is probably broken */
      if (err <= 0)
	return 1;
    }

  return (SSL_get_shutdown (ssl) & SSL_RECEIVED_SHUTDOWN) ? 1 : 0;
}
#endif

int
read_sock (SSL * ssl, char **recvbuf)
{
  int err;
  int i;
  char *newp;
  char sbuf[DBUFSZ] = { 0 };

  *recvbuf = (char *) calloc (DBUFSZ, sizeof (char));

  i = 1;
  while (1)
    {
      memset (sbuf, '\0', DBUFSZ);
      err = SSL_read (ssl, sbuf, DBUFSZ - 1);
      strcat (*recvbuf, sbuf);

      /* Check if we have reached end of message */
      if (*(*recvbuf + strlen (*recvbuf) - 2) == '\r' && 
          *(*recvbuf + strlen (*recvbuf) - 1) == '\n')
        break;
      /* Allocate more memory for message */
      if (err >= DBUFSZ - 1)
        { 
          i++;
          newp = (char *) calloc (DBUFSZ * i, sizeof (char));
          if (newp == NULL)
            {
              /* Ouch ... */
            }
          else
            {
              strcpy (newp, *recvbuf);
              free (*recvbuf);
              *recvbuf = newp;
            }
        }
      /* Connection is probably broken */
      if (err <= 0)
        return 1;
    }
  return (SSL_get_shutdown (ssl) & SSL_RECEIVED_SHUTDOWN) ? 1 : 0;
}

int
write_sock (SSL * ssl, char *msg)
{
  int thisfar;
  int totalsent;

  totalsent = 0;
  while (totalsent < strlen (msg))
    {
      thisfar = SSL_write (ssl, msg, strlen (msg));
      /* Connection is broken */
      if (thisfar <= 0)
	return 1;
      totalsent += thisfar;
    }
  return 0;
}
