/* ruleparser.c
 *
 *  Copyright (C) 2006,2007 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "ruleparser.h"

/* ruleparser states */
enum states
{
  GLOBAL_SPACE = 0,
  RULES_SPACE,
  ACTION_SPACE,
  AFUNC_SPACE,
  GROUP_SPACE
};

/* Return values for rule & action parser */
enum parse_val
{
  OK = 0,
  ERROR,
  BKSTART,
  BKEND
};

static unsigned int arrange_nodes (struct list_s **rlist, struct list_s **alist, char *errstr);
static struct action *allocate_action_node (struct list_s *node, char *bptr);
static unsigned int __parse_rules (struct list_s **rlist, struct list_s **alist, const char *filename,
                                   int *rulenumber, char *errstr);
static enum parse_val parse_rules_space (struct list_s **rlist, enum states pflag, char *bptr,
              		                 const char *filename, int rulenumber, int line,
                                         int groupread, char *errstr);
static enum parse_val parse_afunc_space (struct action *new_node, char *bptr, const char *filename,
		                         int funcread, int line, char *errstr);
static void copy_string (char *bptr, char *cptr);
static int parse_ignorelist (struct d_array *ignore, char *buffer);
static int do_substr_table (struct action *new_node, char *bptr);
static void initialize_action (struct action *acell);
static int do_rsttable (struct action *new_node, char *bptr, char *errstr);
static unsigned int add_rststring (struct action *new_node, char *rststr, char *errstr);

static int read_ignorelist = 1;
static int action_allocated = 0;
static int old_index = -1;
static char *resetphases = NULL;

/* Wrapper for __parse_rules */
unsigned int
parse_rules (struct list_s **rlist, struct list_s **alist,
             struct d_array *rules, char **errstr)
{
  int rulenumber = 0,
      i;

  /* Allocate space for error message */
  *errstr =  (char *) calloc (BUFSIZE, sizeof (char));
  if (*errstr == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }


  for (i = 0; i < d_array_get_size (rules); i++)
    {
      if (__parse_rules (rlist, alist,
                  (const char *) ((struct rulefiles *) \
                                   rules->element[i]->payload)->fpath,
                                   &rulenumber, *errstr) > 0)
        return 1;
    }

  /* Zero rule files */
  if (i == 0)
    {
      snprintf (*errstr, BUFSIZE,
               "Error: zero rule files; no rules initialized\n");
      return 1;
    }

  /* Adjust rule nodes to point correct action nodes */
  if (arrange_nodes (rlist, alist, *errstr) > 0)
    return 1;

  /* Everything is Ok, free errstr */
  free (*errstr);
  *errstr = NULL;

  return 0;
}

static unsigned int
__parse_rules (struct list_s **rlist, struct list_s **alist,
               const char *filename, int *rulenumber, char *errstr)
{
  FILE *fp;
  int line = 0;
  int rulesread = -1;
  int actionread = -1;
  int funcread = -1;
  int groupread = -1;
  char buffer[BUFSIZE];
  char *bptr;
  enum states pflag = GLOBAL_SPACE;
  struct list_s *new_node = NULL;
  struct action *payload;
  enum parse_val val;

  fp = fopen (filename, "r");
  if (fp == NULL)
    {
      snprintf (errstr, BUFSIZE,
	       "Error: can't open specified rule file (%s).\n",
	       filename);
      return 1;
    }

  printf ("Using %s rule file\n", filename);

  while (fgets (buffer, sizeof (buffer), fp) != NULL)
    {
      line++;

      bptr = &buffer[0];

      while (isspace ((int) *bptr))
        bptr++;

      /* Just jump away from comments and empy lines */
      if ((*bptr == '#') || (*bptr == '\0'))
	continue;

      if ((pflag == RULES_SPACE) || (pflag == ACTION_SPACE))
	{
	  if ((actionread == 0 || rulesread == 0) && *bptr != '{')
	    {
	      snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                        filename, line);
              return 1;
	    }
	  else
	    pflag == RULES_SPACE ? rulesread++ : actionread++;

	  if (*bptr == '{')
	    continue;
	  else if (*bptr == '}')
	    {
	      pflag = GLOBAL_SPACE;
	      continue;
	    }
	}

      /* Scan for "rules" indentifier */
      if (!strncmp (bptr, "rules", 5) && pflag == GLOBAL_SPACE)
	{
	  if (rulesread >= 0)
	    {
	      snprintf (errstr, BUFSIZE, "Error: file %s, redefinition of rules at line %d\n",
		       filename, line);
              return 1;
	    }
	  rulesread++;
	  pflag = RULES_SPACE;
	  continue;
	}
      /* Scan for "action" indentifier */
      else if (!strncmp (bptr, "action", 6) && pflag == GLOBAL_SPACE)
	{
	  if (actionread >= 0)
	    {
	      snprintf (errstr, BUFSIZE, "Error: file %s, redefinition of action at line %d\n",
		       filename, line);
              return 1;
	    }
	  actionread++;
	  pflag = ACTION_SPACE;
	  continue;
	}
      else if (pflag == GLOBAL_SPACE)
	{
	  snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                    filename, line);
          return 1;
	}

      /* Start parsing ... */
      if (pflag == RULES_SPACE)
	{
          rulesread++;
          val = parse_rules_space (rlist, pflag, bptr, filename,
                                   *rulenumber, line, groupread, errstr);
	  if (val == OK)
	    (*rulenumber)++;
          else if (val == ERROR)
            return 1;
	  else if (val == BKSTART)
            {
              /* Inside group; rules need to be grouped */
              pflag = GROUP_SPACE;
            }
	}
      else if (pflag == GROUP_SPACE)
        {
          groupread++;
          val = parse_rules_space (rlist, pflag, bptr, filename,
                                   *rulenumber, line, groupread, errstr);
          if (val == BKEND)
            {
              groupread = -1;
              pflag = RULES_SPACE;
            }
          else if (val == OK)
            (*rulenumber)++;
          else if (val == ERROR)
            return 1;
        }
      else if (pflag == ACTION_SPACE)
	{
	  new_node = slist_append_node (alist);
          if (new_node == NULL)
            {
              perror ("malloc() failed");
              exit (1);
            }

	  payload = allocate_action_node (new_node, bptr);
	  actionread++;
	  pflag = AFUNC_SPACE;
	}
      else if (pflag == AFUNC_SPACE)
	{
	  funcread++;
	  val =
	    parse_afunc_space ((struct action *) new_node->payload, bptr,
			       filename, funcread, line, errstr);
	  if (val == BKEND)
	    {
	      funcread = -1;
	      pflag = ACTION_SPACE;
	      action_allocated = 0;
	    }
          else if (val == ERROR)
            return 1;
	}
    }

  if (pflag != GLOBAL_SPACE)
    {
      snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                filename, line);
      return 1;
    }

  fclose (fp);

  return 0;
}

static enum parse_val
parse_rules_space (struct list_s **rlist, enum states pflag, char *bptr, const char *filename,
                   int rulenumber, int line, int groupread, char *errstr)
{
  struct list_s *current = NULL;
  struct rule *payload;
  int i = 0;
  int len;
  int chk = -1;
  int erroffset;
  int offset[4] = { 0 };
  char *rulestr = NULL;
  char *cptr;
  char **cptrptr;
  const char *error;

  /* Adjust pointers to correct strings */
  cptr = &bptr[0];
  while (!isspace ((int) *bptr))
    bptr++;
  *bptr = '\0';
 
  /* We must handle group of rules */ 
  if (strncmp (cptr, "group", 5) == 0)
    return BKSTART; 

  /* Do validation and check status of group block */
  if (pflag == GROUP_SPACE)
    {
      if (*cptr != '{' && groupread == 0)
        {
          snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                    filename, line);
          return ERROR;
        }
      else if (*cptr == '}' && groupread == 1)
        {
          snprintf (errstr, BUFSIZE, "Error: file %s, group ends empty at line %d\n",
                    filename, line);
          return ERROR;
        }
      else if (*cptr == '{')
        return BKSTART;
      else if (*cptr == '}')
        return BKEND;
    }

  /* Add node and allocate space for rule struct */ 
  current = slist_append_node (rlist);
  if (current == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  current->payload = (void *) malloc (sizeof (struct rule));
  if (current->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  /* Mark rule number */
  payload = (struct rule *) current->payload;
  payload->rulenr = rulenumber;

  /* actname indicates action name for rule */
  payload->actname = (char *) calloc (strlen (cptr) + 1, sizeof (char));
  if (payload->actname == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  strncpy (payload->actname, cptr, strlen (cptr));

#if 0
  if (pflag != GROUP_SPACE)
    payload->group = NULL;
  else if (pflag == GROUP_SPACE && groupread == 0)
    {
      payload->group = (struct rulegroup *) malloc (sizeof (struct rulegroup));
    }
#endif

  /* Trim whitespaces and tabs away */
  bptr++;
  while (isspace ((int) *bptr))
    bptr++;

  /* Build string indexing table */
  len = strlen (bptr);
  while (len >= 0)
    {
      if (bptr[len] == '\"')
	{
	  chk++;
          if (len > 0 && chk == 3)
	    chk--;
	  else
	    offset[i++] = len;
	}
      else if ((chk == 1 && !isspace ((int) bptr[len])) ||
               (len == 0 && bptr[len] != '\"'))
	{
	  snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                    filename, line);
	  return ERROR;
	}
      len--;
    }

  /* Allocate space for rule description and regexp and copy
   * corresponding strings to them */
  i = 0;
  cptrptr = &payload->desc;
  while (1)
    {
      cptr = &bptr[offset[i + 1]];
      len = (offset[i]) - offset[i + 1];
      *cptrptr = (char *) calloc (len, sizeof (char));
      if (*cptrptr == NULL)
        {
          perror ("calloc() failed");
          exit (1);
        }

      strncpy (*cptrptr, ++cptr, len - 1);
      if (i)
        break;
      cptrptr = &rulestr;
      i += 2;
    }
 
  /* Do validation for rule node members */
  if (*rulestr == '\0')
    {
      snprintf (errstr, BUFSIZE,
	       "Error: file %s, no search string specified for %s at line %d\n",
	       filename, payload->actname, line);
      return ERROR;
    }
  if (*payload->desc == '\0')
    {
      snprintf (errstr, BUFSIZE,
	       "Error: file %s, no description for rule %s specified at line %d\n",
	       filename, payload->actname, line);
      return ERROR;
    }

  /* Okay, compile regex from rulestr and check that compilation is succesful */
  payload->re = pcre_compile (rulestr, 0, &error, &erroffset, NULL);
  free (rulestr);

  if (payload->re == NULL)
    {
      snprintf (errstr, BUFSIZE,
	       "Error: file %s, failed to compile regular expression for %s rule\n",
	       filename, payload->actname);
      return ERROR;
    }
  
  return OK;
}

static enum parse_val
parse_afunc_space (struct action *new_node, char *bptr, const char *filename, 
                   int funcread, int line, char *errstr)
{
  char *firststr;
  int index = -1;
  struct do_action *ahandler;
  
  if (*bptr != '{' && funcread == 0)
    {
      snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                filename, line);
      return ERROR;
    }
  else if (*bptr == '}' && funcread == 1)
    {
      snprintf (errstr, BUFSIZE, "Error: file %s, action ends empty at line %d\n",
                filename, line);
      return ERROR;
    }
  else if (*bptr == '{')
    return OK;
  else if (*bptr == '}')
    {
      int i;

      /* Mark action phases resetable */
      if (resetphases != NULL && new_node->id != EMPTY)
        {
          int sl;
          int readf = 0;
          char *fptr = NULL;

          /* Mark all actions phases resetable */
          if (strncmp (resetphases, "all", 3) == 0)
            {
              for (i = 0; i < d_array_get_size (new_node->run_action); i++)
                ((struct do_action *)
		  new_node->run_action->element[i]->payload)->allowrst = 1;
            }
          /* Only specified */
          else
            {
              sl = strlen (resetphases);
              for (i = 0; i < sl; i++)
                {
                  if (isdigit ((int) resetphases[i]))
                    {
                      if (readf < 1)
                        {
                          fptr = &resetphases[i];
                          readf = 1;
                        }
                    }
                  else if (isspace ((int) resetphases[i]))
                    {
                      if (readf > 0)
                        {
                          resetphases[i] = '\0';
                          readf = 0;
                          if (atoi (fptr) < d_array_get_size (new_node->run_action))
                            ((struct do_action *)
                             new_node->run_action->element[atoi (fptr)]->payload)->allowrst = 1;
                          else if (atoi (fptr) >= d_array_get_size (new_node->run_action))
                            {
                              snprintf (errstr, BUFSIZE,
                                        "Error: file %s, indexing action phase out of boundaries (%d) with rstphases\n",
                                        filename, atoi(fptr));
                              return ERROR;
                            }
                        }
                    }
                  else
                    {
                      snprintf (errstr, BUFSIZE, "Error: file %s, wrong value for keyword rstphases\n",
                                filename);
                      return ERROR;
                    }
                }
            }
          free (resetphases);
          resetphases = NULL;
        }

      /* Do some validation for run_actions */
      if (new_node->id != EMPTY)
	{
          int rstf = 0;
          struct do_action *ahandler;
	 
          /* Loop through run_action array */ 
          for (i = 0; i < d_array_get_size (new_node->run_action); i++)
	    {
	      ahandler = (struct do_action *) new_node->run_action->element[i]->payload;
	      if (ahandler->event_max == -1)

		{
		  snprintf (errstr, BUFSIZE,
			   "Error: file %s, at line %d, eventcount must be specified for action\n",
			   filename, line);
		  return ERROR;
		}
              else if (ahandler->t_limit == -1)
                {
                  snprintf (errstr, BUFSIZE,
                           "Error: file %s, at line %d, timelimit must be specified for action\n",
                           filename, line);
                  return ERROR;

                }
              else if (ahandler->action == NULL)
                {
                  snprintf (errstr, BUFSIZE,
                           "Error: file %s, at line %d, command must be specified for action\n",
                           filename, line);
                  return ERROR;
                }
	      else if (!read_ignorelist)
		{
		  snprintf (errstr, BUFSIZE,
			   "Error: file %s, at line %d, unterminated ignorelist\n",
			   filename, line);
		  return ERROR;
                }
              else if (ahandler->allowrst)
                rstf = 1;
	    }

          if (new_node->reset != NULL && rstf < 1)
            fprintf (stderr,
                     "Warning: file: %s, rststrings specified, but no rstphases: rststrings has no effect\n",
                     filename);
          else if (new_node->reset == NULL && rstf > 0)
            fprintf (stderr,
                     "Warning: file: %s, rstphases specified, but no rststrings: rstphases has no effect\n",
                     filename);
          if (new_node->reset == NULL && new_node->substrtable[0] > -1)
            fprintf (stderr,
                     "Warning: file: %s, rsttable specified, but no rststrings: rsttable has no effect\n",
                     filename);
	}
      return BKEND;
    }
  else
    {
      /* Just read subsequent string from ignore list */
      if (!read_ignorelist)
        {
          read_ignorelist = parse_ignorelist (new_node->ignore, bptr);
          return OK;
        }

      firststr = &bptr[0];
      while (!isspace ((int) *bptr))
        bptr++;
      *bptr = '\0';
      bptr++;
      while (isspace ((int) *bptr))
        bptr++;

      if (strncmp (firststr, "command", 7) == 0)
	{
	  new_node->id = CMD;

	  sscanf (firststr, "command[%d]", &index);
	  if (index > old_index)
	    action_allocated = 0;

	  if (index < 0 || (action_allocated == 1
	      && ((struct do_action *) new_node->run_action->element[index]->payload)->action
              != NULL) || (action_allocated == 0 && (index > old_index + 1
	      || index < old_index)))
	    {
	      snprintf (errstr, BUFSIZE,
		       "Error: file %s, at line %d, error indexing command\n", filename, line);
	      return ERROR;
	    }

	  if ((new_node->run_action != NULL && d_array_get_size (new_node->run_action) == index)
              || new_node->run_action == NULL)
            initialize_action (new_node);

	  ahandler = (struct do_action *) new_node->run_action->element[index]->payload;

	  /* Allocate space for action string and trim it */
	  ahandler->action =
	    (char *) calloc (strlen (bptr) + 1, sizeof (char));
	  if (ahandler->action == NULL)
	    {
	      perror ("calloc() failed");
	      exit (1);
	    }

          /* Copy string to do_action struct action */
          copy_string (bptr, ahandler->action);

	  if (*ahandler->action == '\0')
	    {
	      snprintf (errstr, BUFSIZE,
		       "Error: file %s, no value for keyword at line %d\n", filename, line);
	      return ERROR;
	    }
	}
      else if (strncmp (firststr, "rststring", 9) == 0)
        {
          int i;

          if (*bptr != '\"')
            {
              snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                        filename, line);
              return ERROR;
            }

          i = strlen (bptr) - 1;
          while (i > 0)
            {
              if (bptr[i] == '\"')
                {
                  bptr[i] = '\0';
                  break;
                }
              if (!isspace (bptr[i]))
                {
                  snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n", 
                            filename, line);
                  return ERROR;
                }
              i--;
            }
   
          /* Add string to reset d_array */
          if (add_rststring (new_node, &bptr[1], errstr) > 0)
            return ERROR;
        }
      else if (strncmp (firststr, "rsttable", 8) == 0)
        {
          int rval;

	  rval = do_rsttable (new_node, bptr, errstr);
          if (rval < 0)
            {
              snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                        filename, line);
              return ERROR;
            }
          else if (rval > 0)
            return ERROR;
        }
      else if (strncmp (firststr, "rstphases", 9) == 0)
        {
          resetphases = (char *) calloc (strlen (bptr) + 1, sizeof (char));
          if (resetphases == NULL)
            {
              perror ("calloc() failed");
              exit (1);
            }
          strncpy (resetphases, bptr, strlen (bptr));
        }
#if 0
      else if (strncmp (firststr, "resetcmd", 8) == 0)
        {
          /* Allocate space for resetcmd and trim it */
          new_node->resetcmd =
            (char *) calloc (strlen (bptr) + 1, sizeof (char));
          if (new_node->resetcmd == NULL)
            {
              perror ("calloc() failed");
              exit (1);
            }
          copy_string (bptr, new_node->resetcmd);
        }
#endif
      else if (strncmp (firststr, "manysources", 10) == 0)
       {
         if (strncmp (bptr, "one", 2) == 0)
           new_node->multisource = 1;
         else if (strncmp (bptr, "all", 3) == 0)
           new_node->multisource = 0;
         else
           {
             snprintf (errstr, BUFSIZE,
                      "Error: file %s, at line %d, manysources value should be one or all\n",
                      filename, line);
             return ERROR;

           }
       }
      else if (strncmp (firststr, "loglevel", 8) == 0)
	{
	  /* Parse logging level */
	  if (strncmp (bptr, "crit", 4) == 0)
	    new_node->loglevel = CRIT;
	  else if (strncmp (bptr, "alert", 5) == 0)
	    new_node->loglevel = ALERT;
	  else if (strncmp (bptr, "info", 4) == 0)
	    new_node->loglevel = INFO;
	  else if (strncmp (bptr, "none", 4) == 0)
	    new_node->loglevel = NONE;
	  else
	    {
	      snprintf (errstr, BUFSIZE,
		       "Error: file %s, at line %d, loglevel value should be crit, alert, info or none\n",
		       filename, line);
	      return ERROR;
	    }
	}
      else if (strncmp (firststr, "logthreshold", 12) == 0)
        {
          /* Parse logging treshold */
          if (strncmp (bptr, "event", 5) == 0)
            new_node->ltvalue = EVENT;
          else if (strncmp (bptr, "always", 6) == 0)
            new_node->ltvalue = ALWAYS;
          else
            {
              snprintf (errstr, BUFSIZE,
                       "Error: file %s, at line %d, logthreshold value should be event or always\n",
                       filename, line);
              return ERROR;
            }
        }
      else if (strncmp (firststr, "eventcount", 9) == 0)
	{
	  sscanf (firststr, "eventcount[%d]", &index);
	  if (index > old_index)
	    action_allocated = 0;

	  if (index < 0 || (action_allocated == 1 &&
	      ((struct do_action*) new_node->run_action->element[index]->payload)->event_max
              != -1) || (action_allocated == 0 && (index > old_index + 1
              || index < old_index)))
	    {
	      snprintf (errstr, BUFSIZE,
		       "Error: file %s, at line %d, error indexing eventcount\n",
		       filename, line);
	      return ERROR;
	    }

          if ((new_node->run_action != NULL && d_array_get_size (new_node->run_action) == index)
              || new_node->run_action == NULL)
            initialize_action (new_node);

	  ahandler = (struct do_action *) new_node->run_action->element[index]->payload;

	  sscanf (bptr, "%d", &ahandler->event_max);
	  if (ahandler->event_max < 0)
	    {
	      snprintf (errstr, BUFSIZE,
		       "Error: file %s, at line %d, eventcount value is less than zero\n",
		       filename, line);
	      return ERROR;
	    }
	}
      else if (strncmp (firststr, "timelimit", 9) == 0)
	{
	  sscanf (firststr, "timelimit[%d]", &index);
	  if (index > old_index)
	    action_allocated = 0;
	  
          if (index < 0 || (action_allocated == 1 &&
	      ((struct do_action *) new_node->run_action->element[index]->payload)->t_limit
              != -1) || (action_allocated == 0 && (index > old_index + 1
	      || index < old_index)))
	    {
	      snprintf (errstr, BUFSIZE,
		       "Error: file %s, at line %d, error indexing timelimit\n",
                       filename, line);
	      return ERROR;
	    }

          if ((new_node->run_action != NULL && d_array_get_size (new_node->run_action) == index)
              || new_node->run_action == NULL)
            initialize_action (new_node);

          ahandler = (struct do_action *) new_node->run_action->element[index]->payload;
	  sscanf (bptr, "%ld", &ahandler->t_limit);
	 
          if (ahandler->t_limit < 0)
	    {
	      snprintf (errstr, BUFSIZE,
		       "Error: file %s, at line %d, timelimit value must be zero or greater\n",
		       filename, line);
	      return ERROR;
	    }
	}
      else if (strncmp (firststr, "ignorelist", 10) == 0)
	{
          if (read_ignorelist)
	    read_ignorelist = parse_ignorelist (new_node->ignore, bptr);
	}
      else if (strncmp (firststr, "usesubstr", 9) == 0)
	{
	  if (do_substr_table (new_node, bptr) < 0)
	    {
	      snprintf (errstr, BUFSIZE, "Error: file %s, parse error at line %d\n",
                        filename, line);
	      return ERROR;
	    }
	}
      else
	{
	  snprintf (errstr, BUFSIZE,
		   "Error: file %s, wrong value for keyword at line %d\n",
                   filename, line);
	  return ERROR;
	}

      old_index = index;
    }

  return OK;
}

static void
copy_string (char *bptr, char *cptr)
{
  int len;
  int i = 0;

  len = strlen (bptr);

  while (*bptr != '\0')
    {
      /* Don't copy newlines or starting/ending "'s */
      if ((*bptr == '\n') || 
           (*bptr == '\"' &&
           (i == 0 || i == len - 2)))
	bptr++;
      else
	*cptr++ = *bptr++;
      i++;
    }
}

static struct action *
allocate_action_node (struct list_s *node, char *bptr)
{
  struct action *payload;
  int i;

  /* Allocate action node */
  node->payload = (void *) malloc (sizeof (struct action));
  if (node->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  payload = (struct action *) node->payload;
  payload->name = (char *) calloc (strlen (bptr) + 1, sizeof (char));
  if (payload->name == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  /* Init action members */
  payload->ignore = d_array_create ();
  if (payload->ignore == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  payload->run_action = NULL;
  payload->reset = NULL;
  payload->resetcmd = NULL;
  payload->id = EMPTY;
  payload->ltvalue = ALWAYS;
  payload->loglevel = NONE;
  payload->multisource = 0;

  strncpy (payload->name, bptr, strlen (bptr));

  /* By default all substrings are used to indentify event 
   * (fill lookup table with 1's) 
   */
  for (i = 0; i < SUBSTRMAX; i++)
    payload->usesubstr[i] = 1;

  /* By default substrtable lookup table is uninitialized
   * (fill lookup table with -1's)
   */
  for (i = 0; i < SUBSTRMAX; i++)
    payload->substrtable[i] = -1;

  return payload;
}

static void
initialize_action (struct action *acell)
{
  int index;
  
  /* We need to create array for events */
  if (acell->run_action == NULL)
    {
      acell->run_action = d_array_create ();
      if (acell->run_action == NULL)
        {
          perror ("malloc() failed");
          exit (1);
        }
    }

  /* Add new element in run_action array */
  index = d_array_add_element (acell->run_action);
  if (index <= -1)
    {
      perror ("d_array_add_element () failed");
      exit (1);
    }

  /* Allocate space for payload */
  acell->run_action->element[index]->payload = (void *)
    malloc (sizeof (struct do_action));
  if (acell->run_action->element[index]->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  /* Initialize run_action members */
  ((struct do_action *) acell->run_action->element[index]->payload)->action = NULL;
  ((struct do_action *) acell->run_action->element[index]->payload)->t_limit = -1;
  ((struct do_action *) acell->run_action->element[index]->payload)->event_max = -1;
  ((struct do_action *) acell->run_action->element[index]->payload)->allowrst = 0;

  action_allocated = 1;
}

/* Arrange rule list nodes aptr pointer to correct action nodes */
static unsigned int
arrange_nodes (struct list_s **rlist, struct list_s **alist, char *errstr)
{
  struct list_s *rcur;
  struct list_s *acur;
  struct rule *rdata;
  struct action *adata;

  rcur = *rlist;

  while (rcur != NULL)
    {
      rdata = (struct rule *) rcur->payload;
      acur = *alist;

      while (acur != NULL)
	{
	  adata = (struct action *) acur->payload;
	  if (!strncmp (rdata->actname, adata->name, strlen (rdata->actname)))
	    {
	      rdata->aptr = (struct action *) acur->payload;
              free (rdata->actname);
              rdata->actname = NULL;
	      break;
            }
          else
            rdata->aptr = NULL;
	  acur = acur->next;
	}

      if (rdata->aptr == NULL)
	{
	  snprintf (errstr, BUFSIZE,
		   "Error: can't find %s action function for rule #%d\n",
		   rdata->actname, rdata->rulenr);
	  return 1;
	}
      rcur = rcur->next;
    }
  return 0;
}

/* This function parse multiple lines for ignore list
 * and allocates/initializes structures for it
 */
static int
parse_ignorelist (struct d_array *ignore, char *buffer)
{
  char *cptr;
  char *element;
  char pname[4096] = { 0 };
  int index;

  cptr = &buffer[0];
  element = &pname[0];

  while (isgraph ((int) *cptr) || *cptr == '/')
    *element++ = *cptr++;
  while (isspace ((int) *cptr))
    cptr++;

  index = d_array_add_element (ignore);
  if (index <= -1)
    {
      perror ("d_array_add_element () failed");
      exit (1);
    }

  ignore->element[index]->payload = (void *) calloc (strlen (pname) + 1, sizeof (char));
  if (ignore->element[index]->payload == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  strcpy ((char *) ignore->element[index]->payload, pname);

  if (*cptr == '\\')
    return 0;
  else
    return 1;
}

static unsigned int
add_rststring (struct action *new_node, char *rststr, char *errstr)
{
  int index;
  int erroffset;
  const char *error;

  /* Create d_array for reset strings */
  if (new_node->reset == NULL)
    {
      new_node->reset = d_array_create ();
      if (new_node->reset == NULL)
        {
          perror ("malloc() failed");
          exit (1);
        }
    }

  index = d_array_add_element (new_node->reset);
  if (index <= -1)
    {
      perror ("d_array_add_element () failed");
      exit (1);
    }

  new_node->reset->element[index]->payload =
    (void *) pcre_compile (rststr, 0, &error, &erroffset, NULL);
  if (new_node->reset->element[index]->payload == NULL)
    {
      snprintf (errstr, BUFSIZE,
               "Error: failed to compile regular expression for %s rststring\n",
               rststr);
      return 1;
    }
  return 0;
}

/* Build lookup table for substrings. Table marks substring
 * indexes which are used to indentify event
 */
static int
do_substr_table (struct action *new_node, char *bptr)
{
  int i;
  int idx;

  for (i = 0; i < SUBSTRMAX; i++)
    new_node->usesubstr[i] = 0;

  for (i = 0; i < strlen (bptr); i++)
    {
      if (bptr[i] != '$' && !isspace ((int) bptr[i])
          && !isdigit ((int) bptr[i]))
	return -1;
      else if (bptr[i] == '$' && !isdigit ((int) bptr[i + 1]))
        return -1;
      else if (bptr[i - 1] != '$' && isdigit ((int) bptr[i]))
        return -1;
 
      /* Found substring index */
      if (bptr[i - 1] == '$' && isdigit ((int) bptr[i]))
	{
	  idx = atoi (&bptr[i]);
	  new_node->usesubstr[idx] = 1;
	}
    }
  return 0;
}

/* Build lookup table for reset substrings */
static int
do_rsttable (struct action *new_node, char *bptr, char *errstr)
{
  int i;
  int k, v;

  for (i = 0; i < strlen (bptr); i++)
    {
      if (isdigit ((int) bptr[i]))
        {
          if (bptr[i + 1] == ':' &&
              isdigit ((int) bptr[i + 2]))
            {
              k = atoi (&bptr[i]);
              v = atoi (&bptr[i + 2]);
              if (k > SUBSTRMAX - 1 || v > SUBSTRMAX - 1)
                {
                  snprintf (errstr, BUFSIZE,
                           "Error: rstphase value exceeds maximum substring value\n");
                  return 1;
                }
              new_node->substrtable[k] = v;
              i += 2;
            }
          else
            return -1;
        }
    }
  return 0;
}

